import argparse
import sys
import re
from pipes import quote
from hashlib import md5
from subprocess import check_output, Popen, PIPE, CalledProcessError, STDOUT

"""
20160624 willem@byte.nl

Parse crontab and add flock to commands, if required.

Caveats: flock requires encapsulation of the command.


# this works
* * * * * ! test -e /ikbestaniet && echo "uitroepteken werkt" > /tmp/succes1

# this doesn't
* * * * * flock -n ~/.testflock ! test -e /ikbestaniet && echo "uitroepteken werkt" > /tmp/succes2
# flock: !: No such file or directory

# but with -c, it does work:
* * * * * flock -n ~/.testflock -c '! test -e /ikbestaniet && echo "uitroepteken werkt" > /tmp/succes2'

# and so does variable substitution:
myvar=banaan
* * * * * flock -n ~/.testflock -c 'echo ik test of ik cron vars kan gebruiken; echo hier dan myvar: $myvar'

"""

CRONCMD = '/usr/bin/crontab'
IGNORE_KEYWORDS = [
    'flock',
    'lockrun',
    'hypernode-image-optimizer',
]


def get_cron(user):
    try:
        return check_output([CRONCMD, '-l', '-u', user], stderr=STDOUT)
    except CalledProcessError as e:
        if e.output.startswith('no crontab for '):
            return ""
        raise


def write_cron(crontab, user):
    p = Popen([CRONCMD, '-u', user, '-'], stdout=PIPE, stdin=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate(input=crontab)
    statuscode = p.wait()
    if statuscode:
        raise RuntimeError("Saving cron, got status %s and output: %s %s" % (statuscode, stdout, stderr))


def validate_timespec(ts):
    if ts.lstrip('@') in ('reboot', 'yearly', 'annually', 'monthly', 'weekly', 'daily', 'midnight', 'hourly'):
        return True

    if re.match('^[,\d\-\*\/\s\w]+$', ts):
        return True

    raise RuntimeError("Invalid crontab timespec found: %s" % ts)


def flockerify_line(line):
    croncmd, _, comment = line.partition('#')
    croncmd = croncmd.strip()

    if not croncmd:
        return line

    for kw in IGNORE_KEYWORDS:
        if kw in line:
            return line

    if re.match('[\S]+\s*=', croncmd):
        # mailto = bla
        return line

    if croncmd[0] == '@':
        # timespec such as @daily
        timespec, cmd = croncmd.split(None, 1)
    else:
        # timespec such as * * * * *
        tokens = croncmd.split(None, 5)
        if len(tokens) != 6:
            raise RuntimeError("Cannot parse crontab line: %s" % line)
        timespec = ' '.join(tokens[:5])
        cmd = tokens[5]

    validate_timespec(timespec)

    if comment:
        comment = ' #' + comment

    lockfile = '~/.%s.lock' % md5(cmd).hexdigest()[:8]

    return ' '.join((timespec, 'flock -n', lockfile, '-c', quote(cmd))) + comment + '\n'


def flockerify(crontab):
    newlines = [flockerify_line(l) for l in crontab.splitlines(True)]
    return ''.join(newlines)


def main():

    parser = argparse.ArgumentParser(description="Add 'flock' to cron commands so they cannot overlap")
    parser.add_argument('--user', help='Which user to modify cron for? Default "app"', default="app")
    parser.add_argument('--write', help='Save changes to disk. Otherwise, print.', action='store_true', default=False)
    parser.add_argument('--file', help=argparse.SUPPRESS)
    args = parser.parse_args()

    if args.file:
        with open(args.file) as fh:
            oldcron = fh.read()
    else:
        oldcron = get_cron(args.user)

    if not oldcron:
        print("No cron found, doing nothing.")
        return 0

    newcron = flockerify(oldcron)

    if args.file:
        print(newcron)
    elif args.write:
        if oldcron != newcron:
            print("Found updates, saving to disk.")
            write_cron(newcron, args.user)
        else:
            print("Couldn't flockerize anything, doing nothing.")
    else:
        print(newcron)


if __name__ == '__main__':
    sys.exit(main())
