import logging
import os
import subprocess
import shlex
import pipes
from socket import gethostname
from kamikaze.common import settings
from time import sleep

logger = logging.getLogger(__name__)
HOSTNAME = gethostname()


def retry_on_exception(max_retries=0, wait_time=0):
    """
    This decorator will retry a function when it raises an exception.
    After #max_retries tries it will re-raise the last exception.
    If a retry is succesful, the original return value is returned.
    Decorator-credits for Rik for conceiving this! --WdG

    :param int max_retries: Amount of retry attempts before giving up
    :param int wait_time: Amount of seconds to wait between attempts
    """
    def decorator(function):
        def wrapper(*f_args, **f_kwargs):
            counter = 0
            while counter <= max_retries:
                counter += 1
                try:
                    return function(*f_args, **f_kwargs)
                except Exception as e:
                    logger.warning("Caught exception {error_type} '{error}' of while running '{function}'!"
                                   .format(error_type=type(e), error=e, function=function))
                    if counter <= max_retries:
                        logger.warning("Retrying! (Attempt %d of %d)" % (counter, max_retries))
                        sleep(wait_time)
                    else:
                        logger.warning("Still failing after %s runs. Propagating exception." % counter)
                        raise
        return wrapper
    return decorator


def run_local_command(cmd, catch_stdout=True, catch_stderr=True):
    """
    Runs a local command on the default shell using Popen().
    Raises RuntimeError when process exit code is not 0.

    Args:
        - cmd (str) :                The command to run
        - catch_stdout (bool) :     By default, stdout is suppressed and returned in a tuple;
                                     Supply False if this is unwanted behaviour.
        - catch_stderr (bool) :   Idem dito for stderr.

    Returns:
        Tuple consisting of:
            - process exit code
            - process stdout (if suppressed, else None)
            - process stderr (if suppressed, else None)
    """
    logger.debug("Running local command: '%s'" % cmd)
    if catch_stdout:
        logger.debug("stdout will be suppressed.")
    else:
        logger.debug("stdout will not be suppressed.")

    if catch_stderr:
        logger.debug("stderr will be suppressed.")
    else:
        logger.debug("stderr will not be suppressed.")

    stdout_arg = subprocess.PIPE if catch_stdout else None
    stderr_arg = subprocess.PIPE if catch_stderr else None

    subprocess_object = subprocess.Popen(cmd,
                                         stdout=stdout_arg,
                                         stderr=stderr_arg,
                                         shell=True)
    logger.debug("Waiting for local process to finish...")
    stdout, stderr = subprocess_object.communicate()

    logger.debug("Process ended.")

    if catch_stdout:
        logger.debug("Last 100 stdout chars")
        logger.debug(stdout[-100:])

    if subprocess_object.returncode != 0:
        logger.critical("Process stopped. Logging stderr output:")
        logger.critical("STDERR: \n %s" % stderr)

        raise RuntimeError("Executing local command failed with exit code %s: %s" % (subprocess_object.returncode, cmd))

    logger.debug("Local process finished without errors.")
    return subprocess_object.returncode, stdout, stderr


@retry_on_exception(max_retries=3)
def run_remote_command(cmd,
                       remote_host,
                       private_key_file=None,
                       pipe_to_process=None,
                       catch_stdout=True,
                       catch_stderr=True):
    """
    Runs a remote command using SSH.
    Has the ability to pipe stdout of remote process to local process.

    Raises RuntimeError when process exit code != 0
    Raises RuntimeError when optional local pipe process exits with code != 0

    Args:
        - cmd (string) :                The command to run remotely
        - remote_host (string) :        FQDN or IP of remote host
        - private_key_file (string) :   Absolute path to private key to use. Defaults to default key in settings.
        - pipe_to_process (string) :    When given, feed stdout of process to stdin of this extr aprocess.

    Returns:
        A tuple consisting of:
        - Exit code of the SSH process
        - stdout of SSH process
        - stderr of SSH process
    """
    if pipe_to_process:
        if not catch_stderr or not catch_stdout:
            raise RuntimeError("Cannot set up a pipe when stdout or stderr is not caught. Check parameters.")

    # Prepare string to prepend log records
    logprefix = "[(%s) --> (%s)] " % (HOSTNAME, remote_host)
    logger.debug(logprefix + "Running remote command: '%s'" % cmd)

    # Set the default private key this way because the ArgumentParser
    # feeds us None when the PK isn't given on the CLI
    if not private_key_file:
        private_key_file = settings.DEFAULT_PRIVATE_SSH_KEY

    logger.debug(logprefix + "Using private key: %s" % private_key_file)

    if not os.path.exists(private_key_file):
        raise RuntimeError("Private key not found")

    env_cmd = "/usr/bin/env -i SSH_AUTH_SOCK=/dev/null"
    ssh_client = "/usr/bin/ssh"
    options = "-T -o UserKnownHostsFile=/dev/null " \
              + "-o StrictHostKeyChecking=no " \
              + "-o ConnectTimeout=1 " \
              + "-o PasswordAuthentication=no " \
              + "-i %s" % pipes.quote(private_key_file)

    user_host = "root@%s" % remote_host

    local_command = "%s %s %s %s %s" % \
                    (env_cmd, ssh_client, options, pipes.quote(user_host), pipes.quote(cmd))

    stdout_arg = subprocess.PIPE if catch_stdout else None
    stderr_arg = subprocess.PIPE if catch_stderr else None

    if pipe_to_process:
        local_command += " | " + pipe_to_process

    # Required to get the 1st non-zero exit code in the pipe chain,
    # see http://unix.stackexchange.com/questions/14270/get-exit-status-of-process-thats-piped-to-another/
    # Alternative would be to set Popen.executable = /bin/bash but then
    # you can't set the -o pipefail (WdG)
    bash_wrapper = "bash -o pipefail -c %s" % pipes.quote(local_command)

    logger.debug(logprefix + "Starting local command: %s" % bash_wrapper)
    local_process = subprocess.Popen(shlex.split(bash_wrapper),
                                     stdout=stdout_arg,
                                     stderr=stderr_arg,
                                     shell=False)

    stdout, stderr = local_process.communicate()
    returncode = local_process.returncode

    if returncode != 0:
        logger.error(logprefix + "Either SSH or remote process stopped. Logging stdout & stderr output")
        logger.error(stdout + stderr)

        raise RuntimeError(logprefix + "Executing remote command failed " +
                           "with exit code %s: %s. See hypernode log for stdout/stderr." %
                           (returncode, cmd))

    logger.debug(logprefix + "SSH process and remote process finished without errors.")
    return returncode, stdout, stderr
