import argparse
import logging
import json
from kamikaze.colorformatter import ColorFormatter
from kamikaze.migration.migrate import migrate
from kamikaze.migration.settings import MAGENTO_DEFAULT_LOCATIONS
from kamikaze.migration.util import get_hypernode_name, check_can_skip_remote_conn, get_app_name
from kamikaze.migration.notifications import send_slack_message

log = logging.getLogger(__name__)


def parse_arguments(args=None):
    hypernode_name = get_hypernode_name()

    parser = argparse.ArgumentParser(
        description="The old hypernode-importer. Please use the more recent hypernode-importer. "
                    "This importer will be deprecated"
    )
    parser.add_argument('--host', help='SSH hostname to fetch site from')
    parser.add_argument('--port', default=22, help='SSH port to fetch site from', type=int)
    parser.add_argument('--user', help='Username used for SSH connection', default='app')
    parser.add_argument('--path', help='Remote path to look for the Magento directory')
    parser.add_argument('--set-default-url', dest="set_default_url", action='store_true',
                        help='Set secure URL of default storefront and admin area to http://%s' % hypernode_name)
    parser.add_argument('--skip-db-import', action='store_true',
                        help='Skip importing the database. Might be handy if you only want to sync files')
    parser.add_argument('--skip-file-import', action='store_true',
                        help='Skip importing the webroot. Might be handy if you only want to sync the db')
    parser.add_argument('--try-copy-additional-dbs', action='store_true',
                        help='Also try to copy additional databases [experimental]. Will look for a wp-config.php and '
                             'tries to sync the database as defined there as well.')
    parser.add_argument('--dry-run', action='store_true', help="Don't really import but check if the source is OK")
    parser.add_argument('--safe-mode', action='store_true', help='Safe mode aborts importing process '
                        'if any special characters are parsed from the remote local.xml')
    parser.add_argument(
        '--skip-autodetect',
        action='store_true',
        help='Do not attempt to find Magento in the default locations if no Magento can be found in the provided path. '
             'The default locations we look in are {}'.format(MAGENTO_DEFAULT_LOCATIONS)
    )
    parser.add_argument(
        '--set-maintenance-on-source',
        action='store_true',
        help='Set the original shop in maintenance mode. You probably want this '
             'if you are moving the original shop instead of only copying it.'
    )

    # path to file that contains path/files that will be excluded during rsync. Hidden option
    parser.add_argument('--protected-file', help="An optional file to pass as a 'protect' filter to rsync to exclude "
                                                 "files during the sync",
                        action='append', metavar='path', default=list())

    parser.add_argument('--import-from-local-filesystem', help="Specify an archive to get the webroot from instead of "
                                                               "syncing the files from the remote host. Allowed "
                                                               "formats: .rar, .tar, .tar.gz, .tar.bz2",
                        metavar='path')

    parser.add_argument('--use-sql-for-db', help="Import the database from a sql dump instead of syncing it from the "
                                                 "remote host. Allowed formats: .sql, .sql.gz", metavar='path')

    parser.add_argument('--dev', help=argparse.SUPPRESS, action='store_true')

    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument('--verbose', '-v', action='store_true', help='Display debug output', default=True)
    verbosity.add_argument('--quiet', '-q', action='store_false', help='Do not display debug output', dest='verbose')

    # for when we want to import without passing arguments as parameters so we don't have to deal with escaping
    parser.add_argument('--use-settings-file', type=str, help="Parse a json file for keys and values that will be used "
                                                              "as arguments for the importer. Can be handy for "
                                                              "scripting purposes. Example content: "
                                                              "{\"path\": \"/data/web/public\"}", required=False)

    parsed_args = parser.parse_args(args)

    if parsed_args.use_settings_file:
        try:
            with open(parsed_args.use_settings_file) as settings_file:
                cli_kwargs = json.load(settings_file)
            for arg, value in cli_kwargs.iteritems():
                setattr(parsed_args, arg, value)
        except (OSError, ValueError):
            log.error("The settings file you specified contains no valid JSON! Skipping..")

    if parsed_args.protected_file and parsed_args.import_from_local_filesystem:
        raise parser.error("Can't use --protected-file with --import-from-local-filesystem, "
                           "we don't use rsync when importing from .zip")

    can_skip_remote_connection = check_can_skip_remote_conn(import_path=parsed_args.import_from_local_filesystem,
                                                            sql_file=parsed_args.use_sql_for_db,
                                                            skip_file_import=parsed_args.skip_file_import,
                                                            skip_db_import=parsed_args.skip_db_import)
    if not all([parsed_args.host, parsed_args.user]) and not can_skip_remote_connection:
        raise parser.error("Please specify --host and --user")

    if any([parsed_args.host, parsed_args.path]) and can_skip_remote_connection:
        # it is no use specifying both a local filesystem, db AND a remote host. would get both from local storage
        # and thus we would have no use for the remote host information because we don't even need to log in.
        raise parser.error("It is no use specifying both a remote host, a zip and an sql."
                           "Specify only a host and no or one of either zip and sql, "
                           "or either no host and a zip and sql.")
    return parsed_args


def configure_logging(debug=True):
    logger = logging.getLogger('kamikaze.migration')
    level = logging.DEBUG if debug else logging.INFO
    logger.setLevel(level)
    ch = logging.StreamHandler()
    formatter = ColorFormatter()
    ch.setFormatter(formatter)
    logger.addHandler(ch)


def main():
    args = parse_arguments()
    configure_logging(args.verbose)
    send_slack_message("App {} used the hypernode-importer-legacy".format(get_app_name()))

    return migrate(
        args.user, args.host, args.port, args.path, args.set_default_url,
        protected_files=args.protected_file,
        import_path=args.import_from_local_filesystem,
        sql_file=args.use_sql_for_db,
        skip_file_import=args.skip_file_import,
        skip_db_import=args.skip_db_import,
        skip_failure_notifications=args.dev,
        dry_run=args.dry_run,
        safe_mode=args.safe_mode,
        set_maintenance=args.set_maintenance_on_source,
        copy_additional_databases=args.try_copy_additional_dbs,
        skip_autodetect=args.skip_autodetect
    )
