import re
import logging
import os
import pipes
import subprocess
from collections import defaultdict
from time import sleep

from kamikaze.migration.files import sync_remote_app_etc_local_xml
from kamikaze.migration.settings import DESTPATH, LOCALXML, MYCNF
from kamikaze.migration.ssh_connection import wrap_command_in_ssh_call
from kamikaze.migration.util import run_local_command, get_random_port, SpecialCharacterException, has_only_safe_chars
import ConfigParser
from contextlib import contextmanager
from xml.etree.cElementTree import ElementTree

log = logging.getLogger(__name__)

SQLDUMP_CLEAN_SED_SCRIPT = "/tmp/hypernode_importer_clean_sql.sed"

WP_CONFIG_PATTERNS = defaultdict(
    lambda: r'(?!)',  # zero-width negative look ahead assertion, always matches nothing
    dbname=r"(?!#).*DB_NAME.*['\"](.*)['\"].*",
    host=r"(?!#).*DB_HOST.*['\"](.*)['\"].*",
    username=r"(?!#).*DB_USER.*['\"](.*)['\"].*",
    password=r"(?!#).*DB_PASSWORD.*['\"](.*)['\"].*",
    table_prefix=r"(?!#).*table_prefix\s*=\s*['\"](.*)['\"].*"
)


def write_sqldump_sed_rules():
    sed_rules = """#!/bin/sed -f
# Clean a SQL dump so it can be imported into the Hypernode database

# If there are definers, change the user to the app user
s/DEFINER=.*[\*]/DEFINER=\`app\`@\`%\`*\/ \/\*/g

# Determines whether InnoDB performs duplicate key checks.
# Import is much faster for large data sets if this check is not performed.
1s/^/\/\*\!40014 SET @OLD_UNIQUE_CHECKS=@@UNIQUE_CHECKS, UNIQUE_CHECKS=0 \*\/;\\n/

# Determines whether the server should check that a referenced table exists when defining a foreign key.
# Due to potential circular references, this check must be turned off for the duration of the import,
# to permit defining foreign keys.
1s/^/\/\*\!40014 SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0 \*\/;\\n/

# Allow importing of data from AUTO_INCREMENT columns. NULL generates the next sequence number.
# Also see http://dev.mysql.com/doc/refman/5.7/en/sql-mode.html
1s/^/\/\*\!40101 SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE='NO_AUTO_VALUE_ON_ZERO' \*\/;\\n/

# Unset the previous three settings after we are done
$s/$/\\n\/\*\!40014 SET UNIQUE_CHECKS=@OLD_UNIQUE_CHECKS \*\/;/
$s/$/\\n\/\*\!40014 SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS \*\/;/
$s/$/\\n\/\*\!40014 SET SQL_MODE=@OLD_SQL_MODE \*\/;/
"""
    with open(SQLDUMP_CLEAN_SED_SCRIPT, 'w') as f:
        f.write(sed_rules)


def stream_sql_dump_command(sqlfile):
    """
    Parse the extension of the sqlfile and read the contents to standard out
    :param sqlfile:
    :return str: a command that will read the sqlfile to standard out
    """
    if sqlfile.endswith('.gz'):
        return "zcat '{}'".format(sqlfile)
    else:
        return "cat '{}'".format(sqlfile)


def import_sql(sqlfile):
    log.info("Import database from SQL file {}".format(sqlfile))
    dbconf = get_database_credentials_from_file()
    dump_sql = stream_sql_dump_command(sqlfile)
    write_sqldump_sed_rules()
    cmd = "{} | sed -f {} | " \
          "pv -ab --name 'Importing SQL dump {}' 2> /dev/stdout | " \
          "mysql --database '{}'".format(dump_sql, SQLDUMP_CLEAN_SED_SCRIPT,
                                         sqlfile, dbconf['dbname'])
    # todo: escape the cmd because the sqlfile path could be garbage.
    # this is ok for now because only friendlies run the hypernode-importer
    run_local_command(cmd)


def skip_tables_args(dbname, table_prefix=''):
    tables_to_skip = [
        'log_url',
        'log_url_info',
        'log_customer',
        'log_quote',
        'report_event',
        'dataflow_batch_export',
        'dataflow_batch_import',
        'core_file_storage',
        'log_visitor_info',
        'log_visitor',
        'ecomdev_stock_import_tmp',
        'innodb_monitor'
    ]

    def ignore_table_flag(table_name):
        return "--ignore-table=%s.%s%s" % (dbname, table_prefix, table_name)

    return " ".join(map(ignore_table_flag, tables_to_skip))


def mysqldump_command(dbconf, host, port, no_data=False):
    dbname = dbconf['dbname']
    user = dbconf['username']
    password = dbconf['password']
    table_prefix = dbconf['table_prefix']

    # dump runs two times, one to copy all table definitions, one to sync data from
    # select tables only
    no_data = ' --no-data' if no_data else ''
    skip_table_arguments = '' if no_data else skip_tables_args(dbname, table_prefix)

    return "mysqldump -h %s -P %s --opt --single-transaction" \
           " --force --insert-ignore --quote-names" \
           " %s %s -u %s -p%s%s" % \
           (host, port, dbname, skip_table_arguments, user, pipes.quote(password), no_data)


def mysqldump_database(host, port, ssh_user=None, ssh_host=None, ssh_port=22, ssh_password=None, no_data=False,
                       safe_mode=False, database_config=None):
    # TODO: fix absolute paths referenced in the database?
    # Parse db credentials with safe_mode which raises an exception if special chars are parsed for safety reasons
    dbconf = get_database_credentials_from_file(filename=database_config, safe_mode=safe_mode)
    cmd = mysqldump_command(dbconf, host, port, no_data=no_data)
    env = os.environ

    if ssh_host:
        cmd, env = wrap_command_in_ssh_call(cmd, env, ssh_user, ssh_host, ssh_port, ssh_password)

    # Remove superfluous warnings:
    # Warning: Using a password on the command line interface can be insecure.
    # http://stackoverflow.com/a/15936384/604515
    cmd += ' 2> >(grep -v "Using a password" 1>&2)'

    sync_type = 'schema' if no_data else 'data'

    # Force stderr of pv to stdout, so it won't be captured
    # by subprocess.check_output (if anything fails)
    cmd += ' | pv -ab --name "Syncing database %s" 2> /dev/stdout' % sync_type

    cmd += ' | mysql --database "%s"' % dbconf['dbname']

    run_local_command(cmd, cwd=DESTPATH, stderr=subprocess.STDOUT, env=env)


def open_tunnel(ssh_user, ssh_host, ssh_port, ssh_password, database_host,
                database_port, forward_port=None, allow_proxy=True):
    forward_port = forward_port or get_random_port()
    ssh_user_host = ssh_user + '@' + ssh_host
    forward_arg = '%s:%s:%s' % (forward_port, database_host, database_port)
    tunnel_command = ['ssh', '-4', '-p', str(ssh_port), ssh_user_host, '-L', forward_arg, '-N']
    if not allow_proxy:
        tunnel_command.append('-oProxyCommand=none')

    env = os.environ
    if ssh_password:
        env = dict(env, SSHPASS=ssh_password)
        tunnel_command = ['/usr/bin/sshpass', '-e', '--'] + tunnel_command

    return subprocess.Popen(tunnel_command, env=env)


@contextmanager
def forward_mysql_database(ssh_user, ssh_host, ssh_port, ssh_password, database_host, database_port, forward_port=None,
                           allow_proxy=True):
    forward_port = forward_port or get_random_port()

    log.info("Tunneling remote mysql server to localhost")

    p1 = open_tunnel(ssh_user, ssh_host, ssh_port, ssh_password,
                     database_host, database_port, forward_port,
                     allow_proxy=allow_proxy)

    try:
        sleep(1)    # give the subprocess some time to establish the tunnel
        yield forward_port
    finally:
        p1.kill()


def execute_mysql(query, host=None, port=None, username=None, password=None, ssh_user=None, ssh_host=None,
                  ssh_port=22, ssh_password=None):
    # NOTE: this allows for terrible SQL injections, but since this script is only
    # run by friendlies we'll make an exception
    cmd = ['mysql']
    if host:
        cmd.extend(['-h', host])
    if port:
        cmd.extend(['-P', str(port)])
    if username:
        cmd.extend(['-u', username])
    if password:
        cmd.append('-p%s' % password)
    env = os.environ
    if ssh_host:
        # Watch out with this function, it runs a mysql command on the remote host
        # with the database credentials as arguments. potentially the password could
        # break out of the shell escaping and execute '; rm all the things'
        # You should probably only run this function with --safe-mode when using
        # external input (like for running the dry-run with the Hyperform)
        cmd, env = wrap_command_in_ssh_call(
            "{}".format(' '.join(cmd)) + ' --execute "{}"'.format(query),
            env, ssh_user, ssh_host, ssh_port, ssh_password
        )
    else:
        cmd.extend(
            ['--execute', query]
        )
    run_local_command(cmd, stderr=subprocess.STDOUT, env=env if ssh_host else None)


def assert_database_accessible(host='127.0.0.1', port=3306, username=None, password=None, ssh_user=None, ssh_host=None,
                               ssh_port=22, ssh_password=None):
    log.info("Verify that the database @ %s:%s is accessible" % (host, port))
    try:
        execute_mysql('select 1 limit 0;', host=host, port=port, username=username, password=password,
                      ssh_user=ssh_user, ssh_host=ssh_host, ssh_port=ssh_port, ssh_password=ssh_password)
    except subprocess.CalledProcessError:
        raise CannotAccessDatabaseError(
            "Can't access the database. Is it running? Is the password correct?"
        )


def sync_database_through_tunnel(ssh_user, ssh_host, ssh_port, ssh_password, database_config=None, dry_run=False,
                                 safe_mode=True, allow_proxy=True):
    """
    Forward the mysql server specified in the local.xml through the remote
    server to the localhost and mysqldump the db
    into the hypernode local database
    :param str ssh_user: user for the remote ssh server
    :param str ssh_host: host of the remote ssh server
    :param int ssh_port: port of the remote ssh server
    :param str ssh_password: password if we have one
    :param str database_config: where the remote db settings xml file is
    :param bool dry_run: just check if we can access the db, don't really sync it
    :param bool safe_mode: Whether or not to accept potentially risky special characters
                           inside dbhost, password, dbname or dbuser
    :param bool allow_proxy: If false, overwrite any ProxyCommands in the configured SSH config with none
    :return:
    """
    # Parse db credentials with safe_mode which raises an exception if special chars are parsed for safety reasons
    dbconf = get_database_credentials_from_file(filename=database_config, safe_mode=safe_mode)
    with forward_mysql_database(ssh_user, ssh_host, ssh_port, ssh_password, dbconf['host'], dbconf['port'],
                                allow_proxy=allow_proxy) as fwd_port:
        # check if we can actually log in to the forwarded database
        assert_database_accessible('127.0.0.1', fwd_port, dbconf['username'], dbconf['password'])
        if not dry_run:
            # first get the database structure
            mysqldump_database(
                '127.0.0.1', fwd_port, no_data=True, safe_mode=safe_mode, database_config=database_config
            )
            # then dump the data, leaving some unnecessary tables out)
            mysqldump_database(
                '127.0.0.1', fwd_port, no_data=False, safe_mode=safe_mode, database_config=database_config
            )


def sync_database_through_ssh_pipe(ssh_user, ssh_host, ssh_port, ssh_password, database_config=None, dry_run=False,
                                   safe_mode=True):
    log.info("Copying database from remote site using mysqldump on source host")
    # Parse db credentials with safe_mode which raises an exception if special chars are parsed for safety reasons
    dbconf = get_database_credentials_from_file(filename=database_config, safe_mode=safe_mode)
    assert_database_accessible(dbconf['host'], dbconf['port'], dbconf['username'], dbconf['password'],
                               ssh_user=ssh_user, ssh_port=ssh_port, ssh_host=ssh_host, ssh_password=ssh_password)
    if not dry_run:
        # first get the database structure
        mysqldump_database(
            dbconf['host'], dbconf['port'], ssh_user, ssh_host, ssh_port, ssh_password,
            no_data=True, database_config=database_config
        )
        # then dump the data, leaving some unnecessary tables out)
        mysqldump_database(
            dbconf['host'], dbconf['port'], ssh_user, ssh_host, ssh_port, ssh_password,
            no_data=False, database_config=database_config
        )


def try_sync_database_through_tunnel(ssh_user, ssh_host, ssh_port, ssh_password, database_config=None, dry_run=False,
                                     safe_mode=True):
    """
    Try to sync the database through an SSH tunnel forwarding the remote database port
    If we can not sync with the SSH config on disk, try other configurations as well
    (like disabling any potential proxies)
    :param str ssh_user: SSH user
    :param str ssh_host: SSH host
    :param int ssh_port: SSH port
    :param str ssh_password: ssh password
    :param str database_config: database config
    :param bool dry_run: whether or not to really import or just test if we can
    :param bool safe_mode: error out if we encounter suspicious characters
    :return None:
    """
    try:
        sync_database_through_tunnel(ssh_user, ssh_host, ssh_port, ssh_password, database_config=database_config,
                                     dry_run=dry_run, safe_mode=safe_mode)
    except (CannotAccessDatabaseError, IOError, subprocess.CalledProcessError):
        sync_database_through_tunnel(ssh_user, ssh_host, ssh_port, ssh_password, database_config=database_config,
                                     dry_run=dry_run, safe_mode=safe_mode, allow_proxy=False)


def sync_database(ssh_user, ssh_host, ssh_port, ssh_password, path=None, dry_run=False,
                  safe_mode=False, database_config=None):
    try:
        try_sync_database_through_tunnel(ssh_user, ssh_host, ssh_port, ssh_password, database_config=database_config,
                                         dry_run=dry_run, safe_mode=safe_mode)
    except (CannotAccessDatabaseError, IOError):
        try:
            log.info("Failed to connect to forwarded database (the source host might have disabled TcpForwarding)")
            log.info("Will now try synchronising using fallback method (remote mysqldump)")
            sync_database_through_ssh_pipe(ssh_user, ssh_host, ssh_port, ssh_password, database_config=database_config,
                                           dry_run=dry_run, safe_mode=safe_mode)
        except (CannotAccessDatabaseError, IOError, subprocess.CalledProcessError):
            if not database_config:
                # try to copy the database credentials from the remote site. could be different from what is on our
                # local filesystem (or there is no /data/web/public/local.xml yet)
                log.info("Failed to access database on remote machine credentials from local filesystem.\n"
                         "Will attempt to get credentials from remote system and try again.")
                # todo: differentiate between wordpress sync_databases here, if wordpress don't get remote app
                # local xml but wp-config.php instead
                database_config = sync_remote_app_etc_local_xml(ssh_user, ssh_host, ssh_port, ssh_password, path)
                try:
                    sync_database(ssh_user, ssh_host, ssh_port, ssh_password, database_config=database_config,
                                  dry_run=dry_run, safe_mode=safe_mode)
                finally:
                    try:
                        # clean up temporary local.xml file synced from remote
                        os.remove(database_config)
                    except OSError:
                        pass
                return
            else:
                log.info("Failed to access database on remote machine with credentials from remote filesystem.\n")
                log.error("Out of Mysql connection methods to try!")
                raise

    log.info("Database looks good!")


def configure_mysql(database_config=None):
    log.info("Configuring database and user")
    db_config = get_database_credentials_from_file(filename=database_config)
    create_database(db_config['dbname'])
    create_database_user(db_config['username'], db_config['password'], db_config['dbname'])

    # only update app user's .my.cnf when the source magento database user is named app
    if db_config['username'] == 'app':
        update_my_cnf(db_config['username'], db_config['password'])


def update_my_cnf(username, password):
    parser = ConfigParser.SafeConfigParser()
    parser.read(MYCNF)
    parser.set('client', 'user', username)
    parser.set('client', 'password', password)
    parser.write(open(MYCNF, 'w'))


def create_database_user(username, password, dbname):
    execute_mysql("GRANT ALL PRIVILEGES ON `%s`.* TO '%s'@'%%' IDENTIFIED BY '%s'" % (dbname, username, password))


def create_database(dbname):
    execute_mysql('DROP DATABASE IF EXISTS `%s`' % dbname)
    execute_mysql('CREATE DATABASE `%s`' % dbname)


def check_if_wordpress_database_config(filename=None):
    """
    Check if the file is a wordpress database config
    :param str filename: File to check
    :return bool wordpress_config: True if a wp-config file, false if not
    """
    return filename.endswith('wp-config.php')


def get_database_credentials_from_file(filename=None, safe_mode=False):
    """
    Get a dict with database information from a file
    :param str filename: File to get the database credentials from
    :param bool safe_mode: Error out if the DB credentials contain 'unsafe' characters
    :return dict database_credentials: Dict with database information
    """
    if filename and check_if_wordpress_database_config(filename=filename):
        return get_database_credentials_from_wordpress_config(wp_config=filename, safe_mode=safe_mode)
    else:
        return get_database_credentials_from_magento_xml(xmlpath=filename, safe_mode=safe_mode)


def _split_port_in_wp_config(database_credentials):
    """
    Split the database port from the host or set it to the default
    :param dict database_credentials: Dict with database information
    :return dict database_credentials: Dict with database information with port split from the host
    """
    if database_credentials['host'] and ':' in database_credentials['host']:
        database_credentials['port'] = int(database_credentials['host'].split(':')[1] or 3306)
        database_credentials['host'] = database_credentials['host'].split(':')[0]
    database_credentials['port'] = database_credentials['port'] or 3306
    return database_credentials


def _use_default_wp_table_prefix_if_none_specified(database_credentials):
    """
    Use default wp table prefix if none specified
    :param dict database_credentials: Dict with database information
    :return dict database_credentials: Dict with database information with default table prefix if none specified
    """
    database_credentials['table_prefix'] = database_credentials['table_prefix'] or 'wp_'
    return database_credentials


def _add_wordpress_credentials_type_to_credentials(database_credentials):
    """
    Add wordpress credentails type to database credentials
    :param dict database_credentials: Dict with database information
    :return dict database_credentials: Dict with database information with wp credentials type
    """
    database_credentials['credentials_type'] = 'wordpress'
    return database_credentials


def _parse_wordpress_config_for_database_credentials(wp_config):
    """
    Parse wordpress config for database credentials
    :param str wp_config: path to the wordpress config
    :return dict database_credentials: Dict with database information
    """
    with open(wp_config) as f:
        content = f.read()

    def get_configured(key):
        pattern = WP_CONFIG_PATTERNS[key]
        matches = re.findall(pattern, content or '')
        return matches[-1] if matches else None

    keys_to_get = ('host', 'port', 'username', 'password', 'table_prefix', 'dbname')
    defaults = (
        _use_default_wp_table_prefix_if_none_specified,
        _split_port_in_wp_config,
        _add_wordpress_credentials_type_to_credentials
    )
    database_credentials = reduce(
        lambda creds, func: func(creds), defaults,
        {k: get_configured(k) for k in keys_to_get}
    )
    return database_credentials


def get_database_credentials_from_wordpress_config(wp_config, safe_mode=False):
    """
    :param str wp_config: File to get the database credentials from
    :param bool safe_mode: Error out if the DB credentials contain 'unsafe' characters
    :return dict database_credentials: Dict with database information
    """

    database_credentials = _parse_wordpress_config_for_database_credentials(wp_config=wp_config)
    if any(map(lambda v: v is None, database_credentials.values())):
        raise RuntimeError("Failed to parse database credentials from {}. You can try again without "
                           "--try-copy-additional-dbs to skip importing the Wordpress DB".format(wp_config))
    if safe_mode and not all(map(has_only_safe_chars, database_credentials.values())):
        raise SpecialCharacterException("Special characters detected in wordpress dbname/username/host/password. "
                                        "Please use hypernode-importer3 which does support special characters.")
    return database_credentials


def get_database_credentials_from_magento_xml(xmlpath=None, safe_mode=False):
    """
    Parse and return database credentials from <xmlpath>/local.xml
    :param str xmlpath: Path to local.xml, i.e.: /app/etc/local.xml
    :param bool safe_mode: Whether or not to accept potentially risky special characters
                           inside dbhost, password, dbname or dbuser
    :return dict database_credentials:
    """
    xmlpath = xmlpath or LOCALXML
    tree = ElementTree(file=xmlpath)

    # for reference, http://eli.thegreenplace.net/2012/03/15/processing-xml-in-python-with-elementtree
    # cheatsheet: https://gist.github.com/LeCoupa/8c305ec8c713aad07b14
    dbconn = tree.find('.//global/resources/default_setup/connection')
    if not dbconn:
        raise RuntimeError("Could not find database config in %s" % xmlpath)

    try:
        host, _, port = dbconn.find('./host').text.partition(':')
        username = dbconn.find('./username').text
        password = dbconn.find('./password').text
        dbname = dbconn.find('./dbname').text
    except AttributeError:
        raise RuntimeError("Could not find database config in %s" % xmlpath)

    try:
        table_prefix = tree.find('.//global/resources/db/table_prefix').text or ''
    except AttributeError:
        table_prefix = ''

    try:
        port = int(port)
    except (ValueError, TypeError):
        port = 3306

    if safe_mode and not all(map(has_only_safe_chars, (port, username, password, dbname, host))):
        raise SpecialCharacterException("Special characters detected in dbname/username/host/password. Please use"
                                        "hypernode-importer3 which does support special characters.")

    return dict(
        credentials_type='magento',
        host=host,
        port=port,
        username=username,
        password=password,
        dbname=dbname,
        table_prefix=table_prefix
    )


def configure_local_database(database_config=None, safe_mode=False):
    if database_config and check_if_wordpress_database_config(filename=database_config):
        log.info("Fixing database configuration in imported wp-config.php")
    else:
        log.info('Fixing database configuration in imported local.xml Magento')

    with open(database_config or LOCALXML, 'r+b') as local_database_config:
        contents = local_database_config.read()
        old_db_host = get_database_credentials_from_file(filename=database_config, safe_mode=safe_mode)['host']
        contents = contents.replace(old_db_host, 'mysqlmaster')
        local_database_config.seek(0)
        local_database_config.truncate()
        local_database_config.write(contents)


class CannotAccessDatabaseError(RuntimeError):
    pass
