"""
When an expected exception is encountered in the code the global return_code
set. Various exceptions can cause the same return_code, these exit codes
map to a set of errors that can be displayed to the user to inform them of
what they are doing wrong.

For example: the hyperform should not parse the importer output but instead just
look at the exit code and use that to decide what type of error to display.

0    - No error, command succeeded
1    - Catchall for general errors
150  - User aborted the importer with a keyboard interrupt
151  - Circular import was attempted. Can not import the same Hypernode to itself
152  - Authentication failed. Could not log in with either key auth or the provided password
153  - Unsupported special characters in SSH password while running in safe mode
154  - Key authentication failed and no alternative password was provided
155  - No Magento installation found on remote host
156  - Can not access the database defined in app/etc/local.xml
157  - Unsupported special characters in database credentials while running in safe mode
158  - No file copying tool on the remote host. Please install rsync
"""
from collections import defaultdict

NOERROR_RC = 0
CATCHALL_RC = 1
USER_ABORTED_RC = 150
CIRCULAR_IMPORT_RC = 151
SSH_AUTH_FAIL_RC = 152
SSH_SPECIAL_CHAR_RC = 153
SSH_NO_PASS_RC = 154
NO_MAG_RC = 155
NO_DB_ACCESS_RC = 156
DB_SPECIAL_CHAR_RC = 157
NO_FILE_COPYING_TOOL_RC = 158
MAG_2_NOT_SUPPORTED_RC = 159
SSH_TIMEOUT_RC = 160
REDIS_FLUSH_ERR_RC = 161
MAGE_FLUSH_ERR_RC = 162
NO_SUCH_DNS_RC = 163
NO_REMOTE_DB_ACCESS_RC = 164
SSH_NO_ROUTE_RC = 165


IMPORTER_ERROR_MESSAGES = defaultdict(lambda: "of an uncaught error",)
IMPORTER_ERROR_MESSAGES.update({
    NOERROR_RC: None,
    USER_ABORTED_RC: 'the importer was killed',
    CIRCULAR_IMPORT_RC: 'you can not import a hypernode to itself',
    SSH_AUTH_FAIL_RC: 'the ssh credentials were wrong (key + password)',
    SSH_SPECIAL_CHAR_RC: "there were specials chars in ssh password, our automation can not handle this",
    SSH_NO_PASS_RC: "the ssh key was wrong. no password provided.",
    SSH_TIMEOUT_RC: "we could not connect to the provided SSH host (timed out)",
    SSH_NO_ROUTE_RC: "we could not connect to the provided SSH host (no route to host)",
    NO_MAG_RC: "we could find no magento installation",
    NO_DB_ACCESS_RC: "we can not connect to the local database with the found credentials",
    DB_SPECIAL_CHAR_RC: "there were special chars in the db password, our automation can not handle this",
    NO_FILE_COPYING_TOOL_RC: "rsync was not installed on the other server",
    MAG_2_NOT_SUPPORTED_RC: "the shop was magento 2 and this is not supported by the importer yet",
    REDIS_FLUSH_ERR_RC: "we failed to flush the Redis cache, the rest of the import might still have succeeded. "
                        "This was one of the last steps of the importer. You can try to flush it manually with "
                        "'redis-cli flushall'. Also make sure to flush your Magento cache with 'magerun cache:flush'",
    MAGE_FLUSH_ERR_RC: "we failed to flush the Magento cache, the rest of the import might still have succeeded. "
                       "Flushing the cache is the last step of the importer. You can try to flush it manually with "
                       "'magerun cache:flush'",
    NO_SUCH_DNS_RC: "the provided host can not be resolved. Is the DNS correct? Name or service not known.",
    NO_REMOTE_DB_ACCESS_RC: "we can not access the remote database with the found credentials in app/etc/local.xml"
})
