# encoding: UTF-8
import logging
import os.path
import shlex
from pipes import quote
from subprocess import CalledProcessError
from uuid import uuid4
from shutil import rmtree
from distutils.dir_util import copy_tree

from kamikaze.migration.settings import DESTPATH, NGINXPATH
from kamikaze.migration.util import run_local_command, shellify, find_closest_path

log = logging.getLogger(__name__)

RSYNC_IO_ERROR = 23
RSYNC_FILE_VANISHED = 24


def find_magento_path(directory):
    """
    we identify a Magento 1 installation by finding a directory that looks like:
    ├── app
    ├── lib
    └── cron.php
    """

    webroot_identifying_dirs = ['app', 'lib']
    webroot_identifying_files = ['cron.php']
    app_dir_identifying_files = ['Mage.php']

    webroot = find_closest_path(directory, directories=webroot_identifying_dirs, files=webroot_identifying_files)

    if not webroot:
        raise RuntimeError("Can't find Magento webroot! No directory containing {} and {}".format(
            ', '.join(webroot_identifying_dirs),
            ', '.join(app_dir_identifying_files),
        ))

    if webroot + '/app' != find_closest_path(webroot, files=['Mage.php']):
        raise RuntimeError("Can't find a Mage.php in the app directory in the webroot {}!"
                           "Are you sure this is a complete Magento installation?")

    return webroot


def import_from_filesystem(path_to_directory_or_archive):
    """
    Import a Magento shop from the local filesystem. Takes a path to either a directory or archive.
    If the path is an archive (like .zip), unpack it and search there.
    :param str path_to_directory_or_archive: path to the directory or archive to search for the Magento installation in
    :return None:
    """
    if os.path.isdir(path_to_directory_or_archive):
        import_directory(path_to_directory_or_archive)
    else:
        import_archive(path_to_directory_or_archive)


def unarchive_command(archive_path, unarchive_dir):
    """
    Parse the extension of the archive and unpack it to the unarchive path
    :param str archive_path: path to the archive i.e. /tmp/magento.zip
    :param str unarchive_dir: dir to unarchive the archive to

    :return str: a command that will unarchive the archive to the unarchive dir
    """
    if archive_path.endswith('.zip'):
        return "/usr/bin/unzip -q '{}' -d {}".format(archive_path, unarchive_dir)
    elif archive_path.endswith('.rar'):
        return "/usr/bin/unrar e '{}' {}".format(archive_path, unarchive_dir)
    elif archive_path.endswith('.tar.gz') or archive_path.endswith('.tgz'):
        return "/bin/tar -xzf '{}' -C {}".format(archive_path, unarchive_dir)
    elif archive_path.endswith('.tar.bz2'):
        return "/bin/tar -xjf '{}' -C {}".format(archive_path, unarchive_dir)
    elif archive_path.endswith('.tar'):
        return "/bin/tar -xf '{}' -C {}".format(archive_path, unarchive_dir)
    raise RuntimeError("The hypernode importer does not know how to unpack this sort of file!")


def import_archive(archive_path):
    """
    Unpack an archive and try to import a Magento shop from the extracted result
    :param str archive_path: path to the archived (.zip or other) file containing the webroot
    :return None:
    """
    log.info("Will attempt to import shop from archive {}".format(archive_path))
    unarchive_dir = "/data/web/hypernode-importer-temp-{}".format(uuid4().hex)
    try:
        log.info("Unarchiving to temporary directory {}".format(unarchive_dir))
        cmd = shlex.split(unarchive_command(archive_path, unarchive_dir))
        run_local_command(cmd)

        import_directory(unarchive_dir)
    finally:
        # remove the temporary unzipped directory
        log.info("Cleaning up temporary unpack directory..")
        rmtree(unarchive_dir, ignore_errors=True)


def import_directory(directory):
    """
    Find a Magento installation in the directory and copy it to the webroot.
    :param str directory: The directory to look for the Magento installation in
    :return None:
    """
    magento_path = find_magento_path(directory)
    log.info("Copying contents from {} to {}".format(magento_path, DESTPATH))
    copy_tree(magento_path, DESTPATH)


def rsync_files(user, host, port, password, path, options='', ssh_options='', dst=DESTPATH, protected_files=()):
    src = '%s@%s:%s' % (user, host, quote(path))

    protected = ''
    for protected_file in protected_files:
        protected += ' --filter="P {}"'.format(protected_file)
    protected_filters = map('--filter="P {}"'.format, protected_files)
    excludes = map('--exclude={}'.format, protected_files)
    protected = ' '.join(excludes + protected_filters)

    # Include patterns override exclude patterns
    # --itemize-changes required for progress indicator (pv)
    rsync = "rsync -L --archive --itemize-changes --delete {} {}".format(protected, options)

    cmd = shlex.split(rsync)
    cmd_kwargs = {}

    if password:
        cmd_kwargs["add_to_env"] = {"SSHPASS": password}
        cmd = ["sshpass", "-e", "--"] + cmd

    # Squelch informational warnings
    cmd.extend(['-e', 'ssh {}'
                      '-o StrictHostKeyChecking=no '
                      '-o UserKnownHostsFile=/dev/null '
                      '-o VerifyHostKeyDNS=no '
                      '-o Loglevel=error '
                      '-p {}'.format(ssh_options, port)])
    cmd.append(src)
    cmd.append(dst)

    cmd_line = shellify(cmd) + ' | pv -lab --name "Copying files from remote site" >/dev/null'

    try:
        run_local_command(cmd_line, **cmd_kwargs)
    except CalledProcessError as e:
        if e.returncode not in (RSYNC_IO_ERROR, RSYNC_FILE_VANISHED):
            raise
        log.warn("File transfer: some files could not be transfered.")


def try_rsync_files(user, host, port, password, path, options='', dst=DESTPATH, protected_files=()):
    try:
        rsync_files(
            user, host, port, password, path, options=options, dst=dst, protected_files=protected_files
        )
    except CalledProcessError:
        log.debug(
            "Failed syncing the files. Trying again with different SSH config options (forcing ProxyCommand off)"
        )
        rsync_files(
            user, host, port, password, path, dst=dst, options=options, protected_files=protected_files,
            ssh_options="-oProxyCommand=none "
        )


def sync_data_dir(user, host, port, password, path, protected_files=()):
    options = "--include=/var/extendware/* " \
              "--exclude=/var/tmp " \
              "--exclude=/var/session " \
              "--exclude=/var/log " \
              "--exclude=/var/locks " \
              "--exclude=/var/import " \
              "--exclude=/var/export " \
              "--exclude=/var/importexport " \
              "--exclude=/var/cache " \
              "--exclude=/var/backups " \
              "--exclude=/var/report " \
              "--exclude=/media/tmp " \
              "--exclude=/media/js " \
              "--exclude=/media/css_secure " \
              "--exclude=/media/css " \
              "--exclude=/media/wysiwyg/.thumbs " \
              "--exclude=/tmp " \
              "--exclude=/media/catalog/product/cache "
    source_path = "{}/".format(path.rstrip('/'))
    try_rsync_files(
        user, host, port, password, path=source_path, options=options, dst=DESTPATH,
        protected_files=protected_files
    )


def sync_remote_app_etc_local_xml(user, host, port, password, path):
    xml_file = os.path.join(path, 'app/etc/local.xml')
    temp_file_location = '/tmp/{}'.format(uuid4().hex)
    log.info("Copying remote app/etc/local.xml to temporary file")
    try_rsync_files(user, host, port, password, path=xml_file, dst=temp_file_location)
    return temp_file_location


def sync_nginx_dir(user, host, port, password):
    try_rsync_files(user, host, port, password, path=NGINXPATH, dst=NGINXPATH)


def fix_modman_links():
    if os.path.isdir(os.path.join(DESTPATH, '.modman')):
        log.info("Fixing links of modman modules")
        run_local_command('modman repair'.split(), cwd=DESTPATH)
    else:
        log.debug("Modman dir not found, skipping repair")
