import logging
import os
from contextlib import contextmanager
from xml.etree import ElementTree as ET

from kamikaze.migration.database import execute_mysql, get_database_credentials_from_file
from kamikaze.migration.settings import LOCALXML, DESTPATH
from kamikaze.migration.util import run_remote_command, get_hypernode_name, \
    try_remote_command, RemoteCommandFailed
from kamikaze.utils import suppress

log = logging.getLogger(__name__)

default_redis_cache_backend = """
    <cache>
        <backend>Cm_Cache_Backend_Redis</backend>
        <backend_options>
            <server>redismaster</server>
            <port>6379</port>
            <persistent></persistent>
            <database>{}</database>
            <password></password>
            <force_standalone>0</force_standalone>
            <connect_retries>1</connect_retries>
            <read_timeout>10</read_timeout>
            <automatic_cleaning_factor>0</automatic_cleaning_factor>
            <compress_data>1</compress_data>
            <compress_tags>1</compress_tags>
            <compress_threshold>20480</compress_threshold>
            <compression_lib>gzip</compression_lib>
        </backend_options>
    </cache>
"""


def local_xml_exists():
    return os.path.exists(LOCALXML)


def fix_cache_config(path=LOCALXML, tag="cache", database=0):
    log.info('Fixing cache configuration in Magento (%s)', path)
    tree = ET.parse(path)

    globalconfig = tree.find('./global')

    for cache in globalconfig.findall(tag):
        globalconfig.remove(cache)

    redis_backend = ET.fromstring(default_redis_cache_backend.format(database))
    redis_backend.tag = tag  # set the cache element to the requested tag type

    globalconfig.append(redis_backend)

    tree.write(path, encoding="utf-8", xml_declaration=True)


def get_admin_path(path=LOCALXML):
    tree = ET.parse(path)
    path = tree.find('./admin/routers/adminhtml/args/frontName')
    return path.text if path is not None else 'admin'


def set_default_url(path=LOCALXML):
    """
    Sets the default secure and unsecure URLs in the default scope to the Hypernode hostname, e.g. example.hypernode.io

    :return str: new base url
    """
    dbconf = get_database_credentials_from_file()
    adminpath = get_admin_path(path=path)
    hostname = get_hypernode_name()

    new_base_url = "http://%s/" % hostname
    new_admin_url = new_base_url + adminpath + '/'
    log.info("Setting base URL for store to %s & %s" % (new_base_url, new_admin_url))

    execute_mysql("USE {db[dbname]}; UPDATE {db[table_prefix]}core_config_data SET value = '{new_base_url}' WHERE path IN ('web/secure/base_url', 'web/unsecure/base_url')".format(
        db=dbconf, new_base_url=new_base_url))
    execute_mysql("USE {db[dbname]}; UPDATE {db[table_prefix]}core_config_data SET value = 0 WHERE path = 'admin/url/use_custom'".format(db=dbconf))
    return (new_base_url, new_admin_url)


def has_redis_support():
    return os.path.exists(DESTPATH + "/app/code/community/Cm/Cache/Backend/Redis.php") or \
        os.path.exists(DESTPATH + "/lib/Cm/Cache/Backend/Redis.php")


@contextmanager
def manage_maintenance_if_needed(ssh_conn, path, set_maintenance=False, dry_run=False):
    if set_maintenance:
        if dry_run:
            log.info("Skipping setting maintenance mode, doing a dry drun!")
        else:
            log.info("Setting maintenance mode, source shop will go offline now!")
            set_maintenance_mode_on_source(ssh_conn, path=path)
        yield
        # note: this contextmanager should NOT catch the exception and not run the exit functions before returning,
        # finally is not used here because when something goes wrong after setting the maintenance flag on the
        # source host, we wouldn't want the destination host (the Hypernode) to become available because the sync
        # is probably incomplete. If that happens, fix whatever is wrong that caused the importer to fail and then
        # run it again (managing maintenance mode is idempotent this way).
        if dry_run:
            log.info("Skipping ensuring maintenance mode is unset on new copy, doing a dry drun!")
        else:
            log.info("Ensuring the maintenance flag is removed on the new local copy of the site")
            unset_maintenance_mode_on_destination()
    else:
        log.debug("Not managing maintenance mode.")
        yield


def set_maintenance_mode_on_source(ssh_conn, path):
    remote_command = "touch '{}'".format(
        os.path.join(path, 'maintenance.flag')
    )
    failed_message = "Failed to set maintenance mode on source shop!"
    try_remote_command(ssh_conn, remote_command, failed_message)


def unset_maintenance_mode_on_destination():
    with suppress(OSError):  # todo: replace this with FileNotFoundError when we move to python3
        os.remove(
            os.path.join(DESTPATH, 'maintenance.flag')
        )


def find_remote_magento_directory(ssh_conn, path, version):
    """
    Find a directory that looks like a Magento installation in a path on the remote host. If there are more than one
    the first one returned by find will be used, if there is none an empty string will be returned by
    run_remote_command. If the directory does not exist on the remote machine, the run_remote_command will fail with
    a RemoteCommandFailed (RuntimeError) exception and make this function return None.
    :param obj ssh_conn: paramiko ssh connection
    :param str path: the directory to look for a Magento installation in
    :param int version: Magento version. 1 or 2
    :return str: path to the Magento installation
    """
    log.info("Looking for Magento {} installation".format(version))
    find_magento_commands = {
        1: "find {} -name local.xml 2>&1 | grep 'app/etc/local.xml' "
           "| sed 's/app\/etc\/local.xml//g' | grep -v 'Permission denied\|No such file or directory' -i | "
           "head -n 1".format(path),
        2: "find {} -name static.php 2>&1 | grep -v 'Permission denied\|No such file or directory' -i "
           "| head -n 1 | sed 's/static.php//'".format(path)
    }
    find_magento_command = find_magento_commands[version]
    try:
        found_magento_installation = run_remote_command(ssh_conn, find_magento_command)
        if found_magento_installation:
            magento_installation_path = found_magento_installation.strip()
            log.info("Found Magento {} installation in {}!".format(version, magento_installation_path))
            return magento_installation_path
    except RemoteCommandFailed:
        log.info("Did not find Magento {} installation".format(version))


def find_remote_magento_installation(ssh_conn, path):
    """
    Look for a Magento 1 or 2 installation on the remote host in the provided path
    If no Magento 1 installation is found raise a RuntimeError
    If a Magento 2 installation is found raise a NotImplementedError
    :param obj ssh_conn: paramiko ssh connection
    :param str path: the directory to look for a Magento installation in
    :return str: path to the Magento installation
    """
    log.info("Verifying remote Magento installation..")
    magento_1_path = find_remote_magento_directory(ssh_conn, path, version=1)
    if not magento_1_path:
        magento_2_path = find_remote_magento_directory(ssh_conn, path, version=2)
        if magento_2_path:
            raise NotImplementedError(
                "Importing Magento 2 with the hypernode-importer is not supported yet. Please try hypernode-importer3 "
                "which does support Magento 2 imports."
            )
        raise MagentoNotFound(
            "No Magento 1 installation found in {}".format(path)
        )
    return magento_1_path


class MagentoNotFound(RuntimeError):
    pass
