import fnmatch
import logging
import os
import getpass
import socket
from errno import ETIMEDOUT, EHOSTUNREACH
from pipes import quote
from subprocess import CalledProcessError

from paramiko import SSHException

from kamikaze.migration.error_codes import REDIS_FLUSH_ERR_RC, MAGE_FLUSH_ERR_RC, NO_SUCH_DNS_RC, NO_REMOTE_DB_ACCESS_RC, \
    SSH_NO_ROUTE_RC
from kamikaze.migration.database import configure_mysql, configure_local_database, sync_database, \
    assert_database_accessible, import_sql, CannotAccessDatabaseError
from kamikaze.migration.error_codes import NOERROR_RC, CIRCULAR_IMPORT_RC, SSH_AUTH_FAIL_RC, SSH_SPECIAL_CHAR_RC, \
    SSH_NO_PASS_RC, NO_MAG_RC, NO_DB_ACCESS_RC, DB_SPECIAL_CHAR_RC, CATCHALL_RC, USER_ABORTED_RC, \
    NO_FILE_COPYING_TOOL_RC, MAG_2_NOT_SUPPORTED_RC, SSH_TIMEOUT_RC
from kamikaze.migration.files import fix_modman_links, sync_data_dir, sync_nginx_dir, \
    import_from_filesystem
from kamikaze.migration.magento import fix_cache_config, has_redis_support, set_default_url, \
    manage_maintenance_if_needed, find_remote_magento_installation, MagentoNotFound
from kamikaze.migration.notifications import send_notifications_if_necessary, send_importer_result_to_slack, \
    send_hypernode_importer_success_event
from kamikaze.migration.settings import LOCALXML, FPCXML, NGINXPATH, DESTPATH, MAGENTO_DEFAULT_LOCATIONS
from kamikaze.migration.ssh_connection import get_ssh_connection
from kamikaze.migration.util import RemoteCommandFailed, get_hypernode_name, \
    check_can_skip_remote_conn, try_remote_command, run_remote_command, SpecialCharacterException, \
    has_only_safe_chars, is_vagrant, run_local_command

log = logging.getLogger(__name__)

return_code = NOERROR_RC


def _set_return_code(rc):
    """
    Set the global return code to a specific error
    :param int rc: the return code to set. See error_codes.py
    :return None:
    """
    global return_code
    return_code = rc


def _get_return_code():
    """
    Get the global return code
    :return int rc: the global return code
    """
    global return_code
    return return_code


# check if a production node is not importing from itself
def verify_not_importing_from_self(host):
    if os.path.isfile('/etc/hypernode/is_vagrant'):
        return
    if host == get_hypernode_name():
        _set_return_code(CIRCULAR_IMPORT_RC)
        message = "Looks like you tried importing a production Hypernode from the same production Hypernode!\n"
        "Perhaps you meant to import the production Hypernode into a Vagrant?\nIf so, enter the command in "
        "the shell of the Vagrant, not this production shell."
        raise RuntimeError(message)


def verify_ssh_connection(user, host, port, safe_mode=False):
    password = None
    try:
        ssh = get_ssh_connection(user, host, port=port, safe_mode=safe_mode)
    except SSHException:
        try:
            password = getpass.getpass(
                "Cannot log into %s@%s using SSH key or agent, please provide password: " % (user, host))
            try:
                ssh = get_ssh_connection(user, host, port=port, password=password, safe_mode=safe_mode)
            except SSHException:
                _set_return_code(SSH_AUTH_FAIL_RC)
                raise
            except SpecialCharacterException:
                _set_return_code(SSH_SPECIAL_CHAR_RC)
                raise
        except EOFError:
            _set_return_code(SSH_NO_PASS_RC)
            raise RuntimeError("Key authentication failed and no password was provided. "
                               "Can not log in to remote SSH server.")
    except socket.error as e:
        could_not_connect_message = "Could not connect to the provided SSH address ({}), " \
                                    "is it correct? Maybe there is a firewall?"
        if e.errno == ETIMEDOUT:
            _set_return_code(SSH_TIMEOUT_RC)
            raise RuntimeError(could_not_connect_message.format("timed out"))
        if e.errno == EHOSTUNREACH:
            _set_return_code(SSH_NO_ROUTE_RC)
            raise RuntimeError(could_not_connect_message.format("no route to host"))
        if e.errno == socket.EAI_NONAME or e.errno == socket.EAI_AGAIN:
            _set_return_code(NO_SUCH_DNS_RC)
            raise RuntimeError(
                "Could not resolve the provided SSH address. Is it correct? Does the DNS exist? "
                "Name or service not known. If you just updated your DNS it is possible that the TTL "
                "hasn't expired yet."
            )
        raise
    return ssh, password


def find_magento_in_expected_locations(ssh_conn, path=None, autodetect=True):
    """
    If no path is provided, the importer will look on the remote host in the default directories
    $HOME, /var/www, /var/html and /www. Once the Magento installation is found the function will return and look
    no further.
    Raises a RuntimeError when the Magento installation can not be found in either the path or the default directories
    if no path was specified. The error tells the user to be more specific with --path
    :param obj ssh_conn: paramiko ssh connection
    :param str path: the directory to look for a Magento installation in
    :param bool autodetect: Find magento in default locations if Magento was not found in the provided path
    :return str: path to the Magento installation
    """
    magento_locations = [quote(path)] if path else MAGENTO_DEFAULT_LOCATIONS
    for search_location in magento_locations:
        log.info("Looking for Magento installation in {}".format(path or "the usual locations"))
        try:
            try:
                magento_installation_path = find_remote_magento_installation(ssh_conn, search_location)
                return magento_installation_path
            except NotImplementedError:
                _set_return_code(MAG_2_NOT_SUPPORTED_RC)
                raise
        except MagentoNotFound:
            log.info("Couldn't find Magento installation in {}!".format(search_location))
            if path and autodetect:
                log.info(
                    "Will try to autodetect the Magento installation in known installation directories now.."
                )
                found_magento = find_magento_in_expected_locations(ssh_conn)  # without the path
                raise MagentoNotFound(
                    "Could not find Magento installation in the provided path, "
                    "however an installation was found in a different path. \n"
                    "Perhaps you meant to run the importer with --path {}".format(
                        found_magento
                    )
                )
    _set_return_code(NO_MAG_RC)
    raise MagentoNotFound("Could not find Magento installation on the remote host! "
                          "Specify the path to the Magento installation with --path")


def check_if_remote_path_is_symlink(ssh_conn, path):
    """
    Check if the remote path is a symlink
    :param obj ssh_conn: ssh connection
    :param str path: the path to check for if it is a symlink
    :return bool is_symlink: True or false
    """
    check_remote_path_is_symlink_command = "file '{path}' | grep symbolic".format(path=path)
    try:
        run_remote_command(ssh_conn, check_remote_path_is_symlink_command)
        log.info("Looks like the provided path is a symlink")
        return True
    except RemoteCommandFailed:
        log.debug("Detected that the provided path is not a symlink")
        return False


def resolve_remote_symlink(ssh_conn, path, safe_mode=False):
    """
    If the specified path is a symlink, return the real path
    :param obj ssh_conn: ssh connection
    :param str path: the directory to look for a Magento installation in
    :param bool safe_mode: error out if suspicious characters are encountered
    :return:
    """
    resolve_path_command = "readlink '{path}'".format(path=path)
    try:
        resolved_path = run_remote_command(ssh_conn, resolve_path_command).strip()
        if resolved_path and safe_mode and not has_only_safe_chars(resolved_path):
            raise RuntimeError(
                "The resolved path contains unsafe characters. "
                "If you are sure this path is OK and safe you "
                "can try again without the --safe-mode flag"
            )
        return resolved_path or path
    except RemoteCommandFailed:
        # If for some reason resolving the symlink fails (if it is a symlink),
        # try the import anyway with the original path
        log.info("Failed to resolve symlink, using path as is: {}".format(path))
        return path


def verify_magento_installation(ssh_conn, path=None, safe_mode=False, autodetect=True):
    """
    Verify there is a Magento installation somewhere in the provided path and return the path to the Magento
    installation.
    :param obj ssh_conn: ssh connection
    :param str path: the directory to look for a Magento installation in
    :param bool safe_mode: error out if suspicious characters are encountered
    :param bool autodetect: Find magento in default locations if Magento was not found in the provided path
    :return str: path to the Magento installation
    """
    if path:  # Don't need to check safe chars or symlink realpath if no path specified
        if safe_mode and not has_only_safe_chars(path):
            raise RuntimeError(
                "Path provided contains unsafe characters. "
                "If you are sure this path is OK and safe you "
                "can try again without the --safe-mode flag"
            )
        if check_if_remote_path_is_symlink(ssh_conn, path):
            path = resolve_remote_symlink(ssh_conn, path, safe_mode=safe_mode)
    return find_magento_in_expected_locations(ssh_conn, path=path, autodetect=autodetect)


def verify_remote_file_copying_tool(ssh_conn):
    """
    Check if rsync is installed on the remote host
    :param obj ssh_conn: paramiko ssh connection
    :return:
    """
    log.info("Checking if there is a file copying tool we can use on the remote host")
    check_rsync_is_installed_command = "rsync --help"
    try:
        run_remote_command(ssh_conn, check_rsync_is_installed_command)
    except RemoteCommandFailed:
        _set_return_code(NO_FILE_COPYING_TOOL_RC)
        raise RuntimeError(
            "No file copying tool installed on the remote host! "
            "Please make sure you have rsync installed"
        )


def verify_remote_tools(ssh_conn):
    """
    Check required remote tools are available
    :param obj ssh_conn: paramiko ssh connection
    :return:
    """
    log.info("Checking if the required tools are installed on the remote host")
    verify_remote_file_copying_tool(ssh_conn)


def verify_nginx_configuration(ssh_conn):
    log.info("Looking for an nginx configuration")
    remote_command = "test -d {}".format(NGINXPATH)
    failed_message = "No nginx directory on remote host"
    try_remote_command(ssh_conn, remote_command, failed_message)


def transfer_nginx_configuration(ssh_conn, user, host, port, password):
    try:
        verify_nginx_configuration(ssh_conn)
    except RemoteCommandFailed:
        log.info("Could not find nginx config files on the remote host, skipping..")
        return
    sync_nginx_dir(user, host, port, password)


def transfer_store(ssh_conn, user, host, port, password, path, protected_files=(), import_path=None):
    if import_path:
        # syncing files from local filesystem
        import_from_filesystem(import_path)
    else:
        transfer_nginx_configuration(ssh_conn, user, host, port, password)
        # syncing files from remote host
        sync_data_dir(user, host, port, password, path, protected_files=protected_files)
    fix_modman_links()


def transfer_database(ssh_user, ssh_host, ssh_port, ssh_password, path=None, sql_file=None, dry_run=False,
                      safe_mode=False, database_config=None):
    if not dry_run:
        configure_mysql(database_config=database_config)
        if sql_file:
            import_sql(sql_file)

    if not sql_file:
        sync_database(
            ssh_user, ssh_host, ssh_port, ssh_password, path=path,
            dry_run=dry_run, safe_mode=safe_mode, database_config=database_config
        )

    if not dry_run:
        configure_local_database(database_config=database_config, safe_mode=safe_mode)


def find_wordpress_config(path=None):
    """
    Find wordpress config in the Magento directory. It probably has a DB that we want as well.
    :param str path: Path to look for wp-config.php in. Uses the DESTPATH if None specified.
    :return str wp_config_path: The wp-config.php path or None
    """
    matches = []
    for root, dirnames, filenames in os.walk(path or DESTPATH):
        for filename in fnmatch.filter(filenames, 'wp-config.php'):
            matches.append(os.path.join(root, filename))
    return matches[0] if matches else None


def transfer_databases(ssh_user, ssh_host, ssh_port, ssh_password, path=None, sql_file=None, dry_run=False,
                       safe_mode=False, copy_additional_databases=False):
    """
    Transfer all the databases we can find
    """
    log.info("Running 'transfer_database' for Magento")
    transfer_database(ssh_user, ssh_host, ssh_port, ssh_password, path=path, sql_file=sql_file, dry_run=dry_run,
                      safe_mode=safe_mode)
    # note: currently we can not import additional databases from sql dumps. If you need that, please import those
    # dumps manually. Also we can only import with the information found in the locally synced webroot instead of also
    # trying to get the potentially more up to date db information from the remote host in case the file sync is
    # skipped.
    if copy_additional_databases:
        wp_database_config = find_wordpress_config(path=path)
        if wp_database_config:
            log.info("Found a wordpress configuration file. Will attempt to sync that DB as well.")
            transfer_database(ssh_user, ssh_host, ssh_port, ssh_password, path=path, sql_file=sql_file, dry_run=dry_run,
                              safe_mode=safe_mode, database_config=wp_database_config)
        else:
            log.info("Found no additional databases to sync. Skipping..")


def configure_cache():
    fix_cache_config(LOCALXML)
    if os.path.exists(FPCXML):
        fix_cache_config(FPCXML, tag="fpc", database=1)

    if not has_redis_support():
        log.warning("No Redis support found, cache will fall back to using files backed")


def purge_redis_cache():
    log.info("Purging Redis cache")
    try:
        run_local_command(["redis-cli", "flushall"])
    except CalledProcessError:
        _set_return_code(REDIS_FLUSH_ERR_RC)
        raise RuntimeError("Failed to flush the Redis cache.")


def purge_magento_cache():
    log.info("Purging Magento cache")
    magento_flush_command = ['n98-magerun', 'cache:flush', '--root-dir=%s' % DESTPATH]
    try:
        return run_local_command(magento_flush_command, cwd=DESTPATH)
    except CalledProcessError:
        _set_return_code(MAGE_FLUSH_ERR_RC)
        raise RuntimeError("Failed to flush the Magento cache.")


def purge_cache():
    purge_redis_cache()
    purge_magento_cache()


def migrate(user, host, port, path, default_url,
            protected_files=(),
            import_path=None, sql_file=None,
            skip_file_import=False, skip_db_import=False,
            skip_failure_notifications=False,
            dry_run=False,
            safe_mode=False,
            set_maintenance=False,
            copy_additional_databases=False,
            skip_autodetect=False
            ):
    _set_return_code(NOERROR_RC)
    password = None

    log.warning("This importer will be deprecated please use 'hypernode-importer' instead ")

    try:
        skip_remote_conn = check_can_skip_remote_conn(import_path, sql_file, skip_file_import, skip_db_import)

        if skip_remote_conn:
            ssh_conn, password = None, None
        else:
            verify_not_importing_from_self(host)
            ssh_conn, password = verify_ssh_connection(user, host, port, safe_mode=safe_mode)

        try:
            if not skip_remote_conn:
                verify_remote_tools(ssh_conn)
                path = verify_magento_installation(
                    ssh_conn, path, safe_mode=safe_mode, autodetect=not skip_autodetect
                )
            try:
                assert_database_accessible()
            except CannotAccessDatabaseError:
                _set_return_code(NO_DB_ACCESS_RC)
                raise
            with manage_maintenance_if_needed(ssh_conn, set_maintenance=set_maintenance,
                                              path=path, dry_run=dry_run):
                if not skip_file_import and not dry_run:
                    transfer_store(ssh_conn, user, host, port, password, path=path, protected_files=protected_files,
                                   import_path=import_path)
                if not skip_db_import:
                    try:
                        transfer_databases(user, host, port, password, path=path, sql_file=sql_file, dry_run=dry_run,
                                           safe_mode=safe_mode, copy_additional_databases=copy_additional_databases)
                    except SpecialCharacterException:
                        _set_return_code(DB_SPECIAL_CHAR_RC)
                        raise
                    except CannotAccessDatabaseError:
                        _set_return_code(NO_REMOTE_DB_ACCESS_RC)
                        raise

            if dry_run:
                log.info("\nProfit: looks like we are ready to import!\n")
            else:
                configure_cache()

                new_urls = set_default_url(LOCALXML) if default_url else None

                purge_cache()
                log.info("Profit: import completed!")

                if new_urls:
                    log.info("Test your Hypernode here:\n")
                    for new_url in new_urls:
                        log.info(' - ' + new_url)

                log.info("\nFor next steps, see http://support.hypernode.com/going-live.html\n")
                if not is_vagrant():
                    send_importer_result_to_slack(success=True)
                    send_hypernode_importer_success_event()
        except KeyboardInterrupt:
            _set_return_code(USER_ABORTED_RC)
    except Exception as e:
        # subprocess saves output in the CalledProcessError exc
        output = getattr(e, "output", "")
        error_args = dict(exc=e, user=user, host=host, port=port, path=path,
                          default_url=default_url, password=password,
                          output=output)

        # Return a specific exit code or 1 if the exception was not caught and handled
        if _get_return_code() == NOERROR_RC:
            _set_return_code(CATCHALL_RC)

        if not skip_failure_notifications:
            send_notifications_if_necessary(error_args, _get_return_code())

        log.critical('Aborted: %s' % e)
        if output:
            log.critical("Output:\n\n%s" % output)

    return _get_return_code()
