import os
import sys
import pipes
import smtplib
import socket
from email.mime.text import MIMEText
from traceback import format_tb
import paramiko
import json
import logging
import requests
from datetime import datetime

from dateutil.tz import tzutc

from kamikaze.migration.error_codes import CATCHALL_RC, IMPORTER_ERROR_MESSAGES
from kamikaze.migration.settings import SLACK_URL
from kamikaze.migration.util import is_vagrant
from kamikaze.onboarding import call_event_hook
from kamikaze.utils import get_app_name

log = logging.getLogger(__name__)

NOTIFICATION_IGNORE_EXCEPTIONS = (paramiko.AuthenticationException,)
MAIL_EXCEPTION_SENDER = 'HN Importer <noreply@hypernode.io>'
MAIL_EXCEPTION_RECIPIENT = ['hypernode@byte.nl', 'support@byte.nl']
IMPORTER_RESULT_CHANNELS = ('#hyperspam', '#onboarding-actie')

MAIL_BODY = """
Hypernode importer failed. Output:

%(output)s

*** Exception:

%(exc_type)s: %(exc)s\n
%(traceback)s

*** CLI arguments:

%(cli_args)s

*** Using password: %(used_password)s

*** ENV:

%(env)s
"""


def mail_exception(error):
    error['used_password'] = True if error['password'] else False
    error['env'] = json.dumps(dict(os.environ), indent=4, sort_keys=True)
    error['cli_args'] = ' '.join(pipes.quote(x) for x in sys.argv)

    msg = MIMEText(MAIL_BODY % error)

    msg['Subject'] = "hypernode-importer-legacy crashed on %s" % socket.gethostname()
    msg['From'] = MAIL_EXCEPTION_SENDER
    msg['To'] = ", ".join(MAIL_EXCEPTION_RECIPIENT)

    smtp = smtplib.SMTP('localhost')
    try:
        smtp.sendmail(MAIL_EXCEPTION_SENDER, MAIL_EXCEPTION_RECIPIENT, msg.as_string())
    except smtplib.SMTPException:
        log.error("Import failed and the script failed to send the exception to Byte technicians, if you think "
                  "this is something we should take a look at please copy this output and create a ticket.")


def send_slack_message(message, channel='#hyperspam'):
    try:
        payload = {'text': message, 'channel': channel,
                   'username': "hypernode-kamikaze",
                   'icon_emoji': ':space_invader:'}
        headers = {'Accept': 'text/plain', 'Content-type': 'application/json'}
        requests.post(SLACK_URL, data=json.dumps(payload), headers=headers)
    except:
        # always ignore slackbot exceptions
        pass


def send_importer_result_to_slack(success=False, return_code=CATCHALL_RC):
    app_name = get_app_name()
    app_identifier = 'app {}'.format(app_name) if app_name else 'an app with an unparsable hostname'

    emoticon = ":party:" if success else ":anguished:"
    success_message = "was successfully imported"
    failure_message = "failed to import because {}".format(IMPORTER_ERROR_MESSAGES[return_code])
    result = success_message if success else failure_message
    for channel in IMPORTER_RESULT_CHANNELS:
        send_slack_message(
            "{} A shop on {} {}.".format(emoticon, app_identifier, result),
            channel=channel
        )


def send_notifications_if_necessary(error, return_code):
    if is_vagrant():
        return

    exc_type, exc, traceback = sys.exc_info()

    error['exc'] = exc
    error['exc_type'] = exc_type
    error['traceback'] = "".join(format_tb(traceback))

    if error['exc'].__class__ in NOTIFICATION_IGNORE_EXCEPTIONS:
        return

    mail_exception(error)
    send_importer_result_to_slack(success=False, return_code=return_code)
    send_hypernode_importer_failure_event()


def send_hypernode_importer_success_event():
    events = {'hypernode_importer_success': datetime.utcnow().replace(tzinfo=tzutc()).isoformat()}
    call_event_hook(events)


def send_hypernode_importer_failure_event():
    events = {'hypernode_importer_failure': datetime.utcnow().replace(tzinfo=tzutc()).isoformat()}
    call_event_hook(events)
