import logging
import pipes

import os
import subprocess

import paramiko
from paramiko.ssh_exception import SSHException

from kamikaze.migration.util import get_ssh_config_for_host, run_local_command, SpecialCharacterException, \
    has_only_safe_chars

log = logging.getLogger(__name__)


def wrap_command_in_ssh_call(remote_command, env, ssh_user, ssh_host, ssh_port, ssh_password):
    # This is probably the simplest way. Docs recommend against it, but the
    # alternative involves having the stdout-stream of paramiko and passing
    # it into the stdin of a subprocess proc. Probably involving thread and
    # Queue's or maybe with some generators. Let's see how far we can get
    # with spawning a shell...
    # I guess dangers are mostly in creating an escaping hell, injection
    # vectors, and limited error handling.

    ssh_command = 'ssh '
    remote_command = '-p %s %s@%s %s' % \
                     tuple(map(pipes.quote, (str(ssh_port), ssh_user, ssh_host, remote_command)))
    # if an ssh_password is supplied, let sshpass handle it
    if ssh_password:
        env = dict(os.environ, SSHPASS=ssh_password)
        remote_command = "sshpass -e -- " + ssh_command + remote_command
    else:
        remote_command = ssh_command + '-oPasswordAuthentication=no ' + remote_command
    return remote_command, env


class ParamikoSSHConnection(object):
    @staticmethod
    def make_connection(ssh, user, host, port=22, password=None, proxy=None):
        ssh.connect(
            host, port=port, username=user, password=password,
            key_filename=get_ssh_config_for_host(host, 'identityfile'),
            sock=paramiko.ProxyCommand(proxy) if proxy else None
        )
        return ssh

    def try_get_conn(self, ssh, user, host, port=22, password=None, proxy=None):
        try:
            return self.make_connection(
                ssh, user, host, port=port, password=password, proxy=proxy
            )
        except SSHException:
            if proxy:
                # If we can't connect using the proxy, try again without the proxy
                # Note: if the Hyperform needs to resort to this strategy and it succeeds,
                # it could be that the site is not importable from the eventual trial (with a
                # different IP) if there is a firewall in place. In that case the operator
                # could manually set up a tunnel to the remote server through the hyperform server
                # like: ssh -L 1222:the_remote_server.nl:22 hyperform.hypernode.io and then import
                # from the 2222 port bound on the localhost of the trial you want to import the shop
                # to.
                log.debug(
                    "Paramiko could not connect to the remote host using the specified jumphost\n"
                    "Trying again without the proxy to see if we can connect directly instead. "
                )
                return self.make_connection(
                    ssh, user, host, port=port, password=password
                )
            else:
                # If there was no proxy to begin with, don't catch the error
                raise

    def __init__(self, user, host, port=22, password=None):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        proxycommand = get_ssh_config_for_host(host, 'proxycommand')
        proxy = proxycommand.replace(':22 ', ':{} '.format(int(port))) if proxycommand else None
        self.conn = self.try_get_conn(
            ssh, user, host, port=port, password=password, proxy=proxy
        )

    def exec_command(self, command):
        _, stdout, stderr = self.conn.exec_command(command)
        return stdout.channel.recv_exit_status(), stdout.read(), stderr.read()


class FallbackSSHConnection(object):
    def __init__(self, user, host, port=22, password=None, safe_mode=False):
        """
        Sometimes the remote host is incompatible with Paramiko. In that case we can try creating a
        drop in replacement ssh object and use that to run remote commands over the shell.
        :param str user: The user to connect as
        :param str host: The host to connect to
        :param int port: The port to connect to
        :param str password: The password to use, or None if no password
        :param bool safe_mode: Error out when encountering suspicious characters in the connection details
        """
        input_to_check = [user, host, str(port)]
        if password:
            input_to_check.append(password)
        if safe_mode and not all(map(has_only_safe_chars, input_to_check)):
            raise SpecialCharacterException("Special characters detected in SSH connection details. Please use "
                                            "hypernode-importer3 which does support special characters.")

        self.host = host
        self.user = user
        self.port = port
        self.password = password
        self.check_conn()

    def check_conn(self):
        _, stdout, _ = self.exec_command('echo "testing connection"')
        if stdout.strip() != 'testing connection':
            raise SSHException(stdout)

    def exec_command(self, command):
        """
        Mimics the Paramiko exec_command method that runs a command on the remote host over ssh
        and returns a tuple of exit code, stdout and stderr. The underlying run_local_command uses check_output
        which only returns a bytestring of the stdout or raises an CalledProcessError.
        todo: refactor run_local_command so it returns exit_code, stdout and stderr
        :param str command: the remote shell command to execute
        :return:
        """
        cmd, env = wrap_command_in_ssh_call(
            command,
            env=os.environ,
            ssh_user=self.user,
            ssh_host=self.host,
            ssh_port=self.port,
            ssh_password=self.password
        )

        try:
            exit_code = 0
            output = run_local_command(cmd, stderr=subprocess.STDOUT, env=env)
        except subprocess.CalledProcessError as e:
            output = e.output
            exit_code = 1
        return exit_code, output, ''


def get_ssh_connection(user, host, port=22, password=None, safe_mode=False):
    """
    :param str user: The user to connect as
    :param str host: The host to connect to
    :param int port: The port to connect to
    :param str password: The password to use, or None if no password
    :param bool safe_mode: Error out when encountering suspicious characters in the connection details
    :return: the SSH client
    :rtype: paramiko.SSHClient || kamikaze.migration.SSHClient
    """
    log.debug("Connecting to %s@%s %s password" % (user, host, "using" if password is not None else "without"))
    try:
        log.debug("Attempting to get Paramiko SSH connection")
        ssh = ParamikoSSHConnection(user, host, port=port, password=password)
    except SSHException:
        log.debug("Failed to get Paramiko SSH connection, attempting to create a "
                  "fallback SSH connection using the shell now..")
        ssh = FallbackSSHConnection(user, host, port=port, password=password, safe_mode=safe_mode)
    return ssh
