# encoding: UTF-8
import logging
import os
import pipes
import socket
import subprocess
from os import walk
from itertools import ifilter
from re import match
from paramiko.config import SSHConfig

from kamikaze.migration.settings import SAFE_CHARACTER_REGEX, VAGRANT_FLAG
from kamikaze.utils import get_app_name, suppress

log = logging.getLogger(__name__)


REMOTE_NOT_WRITABLE_MESSAGE = "Cannot upload Hypernode helpers to %s, please make sure that it is writable"


def is_vagrant():
    """
    True if we are running in a Vagrant, False if not
    :return bool running_in_vagrant: True if this is a Vagrant
    """
    return os.path.isfile(VAGRANT_FLAG)


def get_hypernode_name():
    """
    Get the Hypernode hostname, e.g. dummy.hypernode.io

    Assumes the hostname is like "rar1s5-dummy-magweb-xls.nodes.hypernode.io"

    :return: the Hypernode hostname, e.g. dummy.hypernode.io
    :return str: hypernode name like: vdloo.hypernode.io
    """
    app_name = get_app_name()
    if not app_name:
        raise ValueError("Could not determine full Hypernode name based on current hostname {}".format(
            socket.gethostname())
        )
    return "{}.hypernode.{}".format(
        app_name, 'local' if is_vagrant() else 'io'
    )


def find_closest_path(search_path, directories=list(), files=list()):
    """
    Find the highest path in a directory that contains the dirs and/or files specified

    example:

    webroot/
    ├── app
    │   └── Mage.php
    └── subdir
        └── app
            └── Mage.php
            └── etc

    find_closest_path(directories=['app']) <- would return the path: webroot
    find_closest_path(files=['Mage.php']) <- would return the path: webroot/app
    find_closest_path(diretories=['etc'], files=['Mage.php']) <- would return the path: webroot/subdir/app

    :param search_path: path to traverse
    :param directories: find highest directory containing these dirs
    :param files: find highest directory containing these files
    :return:
    """

    def contains_dirs_and_files(sub_directory):
        _, existing_directories, existing_files = sub_directory

        contains_dirs = set(directories).issubset(existing_directories)
        contains_files = set(files).issubset(existing_files)

        return contains_dirs and contains_files

    # find matching directories. os.walk returns a tuple like (path, dirs, files)
    matching_directories = ifilter(contains_dirs_and_files, walk(search_path, topdown=True))

    # return the first matching path that is the highest up
    try:
        closest_path = matching_directories.next()[0]
        return closest_path
    except StopIteration:
        return None


def shellify(cmd_list):
    return ' '.join(pipes.quote(c) for c in cmd_list)


def run_local_command(command, add_to_env=None, **kwargs):
    """
    Locally execute a command
    :param list[str] | str command: The command to execute
    :param dict[str, str] add_to_env: An optional dict containing keys to be merged into the environment of the
    subprocess
    :param dict[str, V] kwargs: kwargs to pass to subprocess.check_output
    :return:
    """
    if add_to_env:
        # check to see if the environment is given, if not, get the current process environment
        # then merge in the extra environment variables
        kwargs['env'] = dict(kwargs.get('env', os.environ), **add_to_env)

    # Always false, we because we use our own shell invocation if command
    # is only a single argument. This is required so we can use bash's
    # '-o pipefail' feature.
    kwargs['shell'] = False

    if isinstance(command, list):
        log.debug("(local)  %s" % shellify(command))
    else:
        log.debug("(local)  %s" % command)
        command = ['/bin/bash', '-o', 'pipefail', '-c', command]

    return subprocess.check_output(command, **kwargs)


def get_ssh_config_for_host(host, attribute):
    """
    Parse the ssh config (if it exists) for an identityfile for the host
    :param str host: host that should be looked up in the ssh config
    :param str attribute: the item to find. e.g. identityfile or proxycommand
    :return str identityfile | None:
    """
    config = SSHConfig()
    with suppress(IOError):
        config.parse(open('/data/web/.ssh/config'))
        return config.lookup(host).get(attribute)


def run_remote_command(ssh, command):
    """
    Run a command on a remote host
    :param obj ssh: An ssh object
    :param str command: The command to execute
    :return: the output of the command

    Could throw an Exception of type RemoteCommandFailed with fields status, stdout and stderr
    """
    log.debug("(remote) %s" % command)
    exit_code, stdout, stderr = ssh.exec_command(command)
    if exit_code != 0:
        raise RemoteCommandFailed(exit_code, stdout, stderr)
    return stdout


def try_remote_command(ssh_conn, remote_command, failed_message):
    try:
        run_remote_command(ssh_conn, remote_command)
    except RemoteCommandFailed:
        log.warning(failed_message)
        raise


def check_can_skip_remote_conn(import_path=None, sql_file=None, skip_file_import=None,
                               skip_db_import=None):
    do_not_need_file_sync = import_path or skip_file_import
    do_not_need_db_sync = sql_file or skip_db_import
    return do_not_need_file_sync and do_not_need_db_sync


class RemoteCommandFailed(RuntimeError):
    """Exception raised if remote command fails"""
    def __init__(self, status, stdout, stderr):
        self.status = status
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        return u"Remote command terminated with status %d: %s%s" % (self.status, self.stdout, self.stderr)


def get_random_port():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(('127.0.0.1', 0))
    _, random_port = sock.getsockname()
    sock.close()
    return random_port


def has_only_safe_chars(text):
    if isinstance(text, int):
        return True
    is_safe = bool(match(SAFE_CHARACTER_REGEX, text))
    if not is_safe:
        log.info("Unsafe characters in {}".format(text))
    return is_safe


class SpecialCharacterException(RuntimeError):
    pass
