import socket
import json
import sys
import logging
from contextlib import contextmanager
from os.path import devnull
from subprocess import check_output

from kamikaze.common.settings import COMMON_INFO_PATH


def request_confirmation(message=None, prompt="Continue? (yes/no) ", func=sys.exit, args=None, kwargs=None):
    """
    Prompt user for confirmation before continuing program execution.

    :param message:
        An optional message to print before showing the prompt.
    :param prompt:
        The prompt to display.
    :param func:
        The function that is called when the user answers NO
    :param args:
        Positional arguments passed to the above function.
    :param kwargs:
        Keyword arguments passed to the above function.
    """
    if args is None and func == sys.exit:
        args = [1]
    elif args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    if message is not None:
        print(message)

    response = None
    while True:
        if response in ('y', 'yes'):
            return
        elif response in ('n', 'no'):
            func(*args, **kwargs)
            return
        elif response is not None:
            print("Please answer yes or no.")
        response = raw_input(prompt).lower()


# todo: remove when we upgrade to python > 3.4
@contextmanager
def suppress(*exceptions):
    try:
        yield
    except exceptions:
        pass


lock_socket = dict()


def get_lock(name=None):
    # Get lock using Linux sockets, slightly modified from:
    # http://stackoverflow.com/a/7758075/604515
    global lock_socket

    name = name or sys.argv[0]

    if name in lock_socket:
        logging.warn("This process already issued a lock with name '{}', program error!".format(name))
        return False

    lock_socket[name] = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    try:
        lock_socket[name].bind('\0' + name)
        return True
    except socket.error:
        return False


def parse_app_name_from_host_name():
    """
    Try parsing the app name from the hostname
    Returns None if hostname can not be parsed
    :return str | None: app_name
    """
    hostname_components = socket.gethostname().split('-')
    if len(hostname_components) > 1:
        return hostname_components[1]


def parse_short_handle_from_host_name():
    """
    Try parsing the shorthandle from the hostname
    Returns None if hostname can not be parsed
    :return str | None: app_name
    """
    hostname_components = socket.gethostname().split('-')
    if len(hostname_components) > 1:
        return hostname_components[0]


def get_app_name():
    """
    Get the app name from the node on which this script is running
    :str return: app_name
    """
    try:
        # try to get the app name from the common.json file, this is the most accurate
        with open(COMMON_INFO_PATH, 'r') as info:
            return json.load(info)['tag']
    except (IOError, ValueError, KeyError):
        # when the common.json file is not available (app user can not access it)
        # or there is no tag in the file try parsing the app name from the hostname
        return parse_app_name_from_host_name()


def is_xenial():
    return 'xenial' in check_output(['lsb_release', '-a'], stderr=open(devnull, 'wb'))
